//
//  Huffman Code 1.0
//  HuffmanDecoder.hpp
//
//  Created by Amittai Aviram on 2 Sep 2018.
//  Copyright © 2018 Amittai Aviram. All rights reserved.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

#pragma once

#include "TreeNode.hpp"
#include <string>
#include <vector>

/**
 * \brief Class to decode a Huffman-encoded text.
 *
 * Given a binary file containing the Huffman encoding of a text, decodes
 * the text and prints it to stdout.  The input file must be in the binary
 * format defined by the HuffmanEncoder class, with necessary metadata
 * at the beginning of the file.
 */
class HuffmanDecoder
{
    public:
        /**
         * \brief Destructor: frees memory allocated for Huffman tree.
         */
        ~HuffmanDecoder();

        /**
         * \brief Decodes the text encoded in the given file.
         *
         * Uses the metadata at the beginning of the file to reconstruct
         * the Huffman code in order to decode the encoded text that follows
         * in the file.
         * \param input_file binary file produced by HuffmanEncoder.
         * \param output_file text file with the results of decoding.
         */
        void readMessageFromFileAndDecode(const std::string& input_file, const std::string& output_file);

        /**
         * \brief Decodes the text encoded in the given file.
         *
         * Uses the metadata at the beginning of the file to reconstruct
         * the Huffman code in order to decode the encoded text that follows
         * in the file. Call std::string& decodedMessage() to get the resulting text.
         * \param input_file binary file produced by HuffmanEncoder.
         */
        void readMessageFromFileAndDecode(const std::string& input_file);

        /**
         * \brief Prints the decoded text to stdout.
         */
        std::string& decodedMessage();

    private:
        void readMessageFromFile(const std::string& file);
        void decodeFromFile();

        std::size_t tree_index;
        TreeNode * tree;
        std::vector<TreeNode *> tree_nodes;
        std::vector<uint64_t> encoded_message;
        std::string decoded_message;
};
